/**

  @file
  
  @brief implementation of the Aho-Corasick pattern matching algorithm.
    
  most of this code was taken from another module by Georgios Portokalidis
  with his permission. 
  
  ----
  
  Fairly Fast Packet Filter
  http://ffpf.sourceforge.net/

  Copyright (c), 2003 - 2004 Georgios Portokalidis, Herbert Bos & Willem de Bruijn
  contact info : wdebruij_AT_users.sourceforge.net

  ----
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifdef __KERNEL__
#include <linux/string.h>
#else
#include <string.h>
#endif

#include "aho-corasick.h"
#include "xalloc.h"
#include "slist.h"

#include <stdio.h>


#ifdef DEBUG
#define debug(x) x
#else
#define debug(x)
#endif

#define FAIL NULL
#define aho_corasick_fail(x) x->fail
#define aho_corasick_goto(x,i) x->transitions[i]
#define aho_corasick_output(x) x->output

static aho_corasick_t M;




/* Does an aho-corasick search, given a 'string' of length 'n'.  If
   we're able to find a match, returns a positive integer and the
   start-end indices of the match.
 */
aho_corasick_int_t
ahocorasick_KeywordTree_search_helper(aho_corasick_t *g,
				      unsigned char *string,
				      size_t n,
				      size_t *out_start, size_t *out_end)
{
	size_t i, j;
	aho_corasick_state_t *state;
	i = 0;
	for(j = 0,state = g->zerostate ; j < n ; j++)
	{
		while( aho_corasick_goto(state,*(string+j)) == FAIL ) 
		{
			state = aho_corasick_fail(state);
		}
		state = aho_corasick_goto(state,*(string+j));
		if ( aho_corasick_output(state) != 0) 
		{
			*out_start = j - aho_corasick_output(state) + 1;
			*out_end = j+1;
			return state->id;
		}
	}
	*out_start = -1;
	*out_end = -1;
	return 0;
}



/* Similar to the first helper function, but tries to return the longest
   match. */
aho_corasick_int_t
ahocorasick_KeywordTree_search_long_helper(aho_corasick_t *g,
					   unsigned char *string,
					   size_t n,
					   size_t *out_start, size_t *out_end)
{
	size_t i, j;
	aho_corasick_state_t *state;

	*out_start = -1;
	*out_end = -1;
	i = 0;
	for(j = 0,state = g->zerostate ; j < n ; j++)
	{

		if (aho_corasick_goto(state, *(string+j)) == FAIL &&
		    *out_end != -1) 
		{
			return state->id;
		}
		while( aho_corasick_goto(state,*(string+j)) == FAIL )
		{
			state = aho_corasick_fail(state);
		}
		state = aho_corasick_goto(state,*(string+j));
		if ( aho_corasick_output(state) != 0) 
		{
			*out_start = j - aho_corasick_output(state) + 1;
			*out_end = j+1;
		}
	}

	/* If we reach the end of the string, we still have to double check if
	   we had a longest match queued up. */
	if (*out_end != -1) {
		return state->id;
	}
	return 0;
}



static void
aho_corasick_free(aho_corasick_state_t *state)
{
	int i;

	for(i = 0; i < AHO_CORASICK_CHARACTERS ;i++)
		if ( aho_corasick_goto(state,i) != FAIL ) 
			aho_corasick_free(aho_corasick_goto(state,i));

/* 	printf("Freeing state %d\n", state->id); /\* debug *\/ */
	xfree(state);
}

void
aho_corasick_init(aho_corasick_t *g)
{
	if (!g)
	{
		M.zerostate = NULL;
		M.newstate = 0;
	}
	else
	{
		g->zerostate = NULL;
		g->newstate = 0;
	}
}

void
aho_corasick_destroy(aho_corasick_t *in)
{
	int i;
	aho_corasick_t *g = in;

	if (!g)
		g = &M;

	for(i = 0; i < AHO_CORASICK_CHARACTERS ;i++)
		if ( aho_corasick_goto(g->zerostate,i) != FAIL && 
		     aho_corasick_goto(g->zerostate,i)->id > 0 )
			aho_corasick_free(aho_corasick_goto(g->zerostate,i));
	/* dyoo: added to free the last node. */
	if (g->zerostate != NULL) 
	{
		xfree(g->zerostate);
		g->zerostate = NULL;
	}
}

int
aho_corasick_maketree(aho_corasick_t *in)
{
	slist_t queue;
	aho_corasick_state_t *state,*s,*r;
	aho_corasick_t *g = in;
	int i;

	if (!g)
		g = &M;

	slist_init(&queue);

	// Set all FAIL transition of 0 state to point to itself
	for(i = 0; i < AHO_CORASICK_CHARACTERS ;i++)
	{
		if ( aho_corasick_goto(g->zerostate,i) == FAIL )
			aho_corasick_goto(g->zerostate,i) = g->zerostate;
		// Construct fail()
		else
		{
			if ( slist_append(&queue,aho_corasick_goto(g->zerostate,i)) < 0 )
				goto fail;
			aho_corasick_fail(aho_corasick_goto(g->zerostate,i)) = g->zerostate;
		}
	}

	// Set fail() for depth > 0
	while( (r = slist_pop_first(&queue)) != NULL )
	{
		for(i = 0; i < AHO_CORASICK_CHARACTERS ;i++)
		{
			if ( (s = aho_corasick_goto(r,i)) == FAIL )
				continue;
			if ( slist_append(&queue,s) < 0 )
				goto fail;
			state = aho_corasick_fail(r);
			while( aho_corasick_goto(state,i) == FAIL )
				state = aho_corasick_fail(state);
			aho_corasick_fail(s) = aho_corasick_goto(state,i);
			debug(printf("Setting f(%u) == %u\n",s->id,aho_corasick_goto(state,i)->id));
			// Join outputs missing
		}
	}

	slist_destroy(&queue,SLIST_LEAVE_DATA);
	return 0;

fail:
	slist_destroy(&queue,SLIST_LEAVE_DATA);
	return -1;
}

int
aho_corasick_addstring(aho_corasick_t *in,unsigned char *string)
{
	aho_corasick_t* g = in;
	aho_corasick_state_t *state,*s = NULL;
	int j = 0;

	// Initialize
	if (!g)
		g = &M;
	if ( g->zerostate == NULL )
	{
		if ( (g->zerostate = xalloc(sizeof(aho_corasick_state_t))) == NULL )
		{
			xfree(g);
			return -1;
		}
/* 		printf("allocating state %d\n", 0); /\* debug *\/ */
		g->newstate = 1;
		g->zerostate->id = 0;
		g->zerostate->depth = 0;
		aho_corasick_output(g->zerostate) = 0;
		memset(g->zerostate->transitions,(int)FAIL,AHO_CORASICK_CHARACTERS*sizeof(aho_corasick_state_t *));
	}

	state = g->zerostate;

	// As long as we have transitions follow them
	while( *(string+j) != '\0' 
			&& (s = aho_corasick_goto(state,*(string+j))) != FAIL )
	{
		state = s;
		++j;
	}

	if ( *(string+j) == '\0' ) {
		/* dyoo: added so that if a keyword ends up in a prefix
		   of another, we still mark that as a match.*/
		aho_corasick_output(s) = j;
		return 0;
	}

	while( *(string+j) != '\0' )
	{
		// Create new state
		if ( (s = xalloc(sizeof(aho_corasick_state_t))) == NULL )
			return -1;
		s->id = g->newstate++;
/* 		printf("allocating state %d\n", s->id); /\* debug *\/ */
		s->depth = state->depth + 1;
		memset(s->transitions,(int)FAIL,AHO_CORASICK_CHARACTERS*sizeof(aho_corasick_state_t *));
		// Create transition
		aho_corasick_goto(state,*(string+j)) = s;
		debug(printf("%u -> %c -> %u\n",state->id,*(string+j),s->id));
		state = s;
		aho_corasick_output(s) = 0;
		++j;
	}

	aho_corasick_output(s) = strlen(string);

	return 0;
}
